import math
from typing import List

from flockai.models.drones.keyboard_controller_drone import KeyboardControlledDrone
from flockai.utils.intensive_thread import IntensiveThread, IntensityLevel
from flockai.utils.string_generator import StringGenerator
from flockai.webots_controllers.flockai_controller import FlockAIController
from controller import InertialUnit, Camera, LED, Emitter, Receiver, GPS, Compass, Gyro, Motor


class KeyboardMavic2DJI(FlockAIController, KeyboardControlledDrone):
    def __init__(self):
        super(KeyboardMavic2DJI, self).__init__()
        self.initialize()

    def initialize(self):
        self._attach_and_enable_devices()
        self._attach_and_enable_motors()
        self.keyboard.enable(self.basic_time_step)
        self.batterySensorEnable(self.basic_time_step)

    def _attach_and_enable_devices(self):
        """
        Set new devices here and enable them if applicable
        :return:
        """
        self.emitter: Emitter = self.getEmitter("emitter")

        self.receiver: Receiver = self.getReceiver("receiver")
        self.receiver.enable(self.receiver_time_step)

        self.camera: Camera = self.getCamera("camera")
        self.camera.enable(self.camera_time_step)

        self.front_left_led: LED = self.getLED("front left led")
        self.front_right_led: LED = self.getLED("front right led")

        self.imu: InertialUnit = self.getInertialUnit("inertial unit")
        self.imu.enable(self.basic_time_step)

        self.gps: GPS = self.getGPS("gps")
        self.gps.enable(self.basic_time_step)

        self.compass: Compass = self.getCompass("compass")
        self.compass.enable(self.basic_time_step)

        self.gyro: Gyro = self.getGyro("gyro")
        self.gyro.enable(self.basic_time_step)

    def _attach_and_enable_motors(self):
        """
        Define the motors that are used on the robot
        :return:
        """
        # Camera angle motors
        self.camera_roll_motor = self.getMotor("camera roll")
        self.camera_pitch_motor = self.getMotor("camera pitch")
        self.camera_yaw_motor = self.getMotor("camera yaw")

        # Propellers
        self.front_left_motor = self.getMotor("front left propeller")
        self.front_right_motor = self.getMotor("front right propeller")
        self.rear_left_motor = self.getMotor("rear left propeller")
        self.rear_right_motor = self.getMotor("rear right propeller")

        """
        Set any initial field values for drone motors
        :return:
        """
        # Set position & velocity for propellers
        propellers: List[Motor] = [
            self.front_left_motor,
            self.front_right_motor,
            self.rear_left_motor,
            self.rear_right_motor
        ]

        for propeller in propellers:
            propeller.setPosition(float('inf'))
            propeller.setVelocity(1.0)

    def send_msg(self, msg):
        """
        Sends a message from the attached emitter device
        :param msg: The message to send
        :return: None
        """
        if self.emitter is None:
            print("There is not emitter device attached")

        # print(f'{self.name}: I am sending {message}')
        self.emitter.send(msg.encode('utf-8'))

    def receive_msgs(self):
        """
        Receive messages on the receiver device
        :return: None
        """
        if self.receiver is None:
            print("There is no receiver device attached")
            return

        while self.receiver.getQueueLength() > 0:
            message_received = self.receiver.getData().decode('utf-8')
            # print(f'{self.name}: I have received this message {message_received}')
            self.receiver.nextPacket()

    def blink_led_lights(self, time):
        """
        Blink led lights every 2 seconds
        :param time: The simulation time. ie self.getTime()
        :return:
        """
        led_on = (time // 1) % 2 == 1
        self.front_left_led.set(led_on)
        self.front_right_led.set(not led_on)

    def run(self):
        # Wait a second before starting
        while self.step(self.basic_time_step) != -1:
            if self.getTime() > 1:
                break

        while self.step(self.basic_time_step) != -1:
            time = self.getTime()

            # Blink lights
            self.blink_led_lights(time)
            # Fly drone
            self.actuate()

            # Get battery data
            # print(self.batterySensorGetValue())

            # Retrieve image data
            # self.save_image_every_sec(time, 2)

            # Send messages
            self.send_message(StringGenerator.get_random_message(4))
            # Receive messages
            self.receive_messages()



    # def save_image_every_sec(self, time, sec):
    #     """
    #     Saves images every X seconds
    #     :param time: The simulation time
    #     :param sec: The timestep in seconds
    #     :return:
    #     """
    #     config = SimulationConfig()
    #     if config.simulation_enabled:
    #         save_image = (time // 1) % sec == 1
    #         if save_image:
    #             self.camera.saveImage(f"{config.get_simulation_image_directory()}test_{time}.png", 50)  # quality in range [1, 100]
    #     else:
    #         print("Note: Data collection is not enabled")

    # @classmethod
    # def enable_data_collection(cls):
    #     config = SimulationConfig()
    #     config.simulation_enabled = True
    #     print("Running Simulation: ", config.get_simulation_id())
    #     config.create_simulation_image_directory()


import abc


class FlockAIController(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def __init__(self):
        raise NotImplementedError

    @abc.abstractmethod
    def run(self):
        raise NotImplementedError




import abc
from controller import Robot


class IRobot(Robot):
    """
    Inherit functionalities provided by the Webots Robot library and declare abstract methods
    """
    def __init__(self):
        super(IRobot).__init__()

    @abc.abstractmethod
    def initialize(self):
        """
        The actions needed to initialize a drone
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _attach_and_enable_devices(self):
        """
        Attach and enable the devices on a drone
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _attach_and_enable_motors(self):
        """
        Attach and enable the motors on a drone
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _set_variables(self):
        """
        Set variables needed for controlling the drone on air
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _set_constants(self):
        """
        Set constants needed for controlling the drone on air
        :return:
        """
        raise NotImplementedError



from typing import List
import math
from controller import Keyboard, Robot, Emitter, Receiver, Camera, InertialUnit, GPS, Compass, Gyro, LED, Motor
from flockai.utility import Graphics, SimulationConfig


class DroneUtils(Robot):
    def __init__(self):
        super(DroneUtils, self).__init__()
        self._initialize_drone()

    def _initialize_drone(self):
        # self.basic_time_step = int(self.getBasicTimeStep())  # 8  # Timestep
        self.receiver_time_step = 1000
        self.camera_time_step = 10000
        # self.name = self.getName()  # Name

        self._enable_attached_devices()  # Enable attached devices
        self._enable_attached_motors()  # Enable attached motors

        self._set_initial_motor_fields()  # Set any initial value to motors

        self._display_keyboard_controls()  # Display controls
        self._set_empirical_constants()  # Set constants
        self._set_variables()  # Set variables






    # def send_message(self, message: str) -> None:


    # def receive_messages(self) -> None:
    #     """
    #     Receive messages on the receiver device
    #     :param receiver: The attached receiver device instance
    #     :param name: Optional for debugging purposes, the name of the robot
    #     :return: None
    #     """
    #     if self.receiver is None:
    #         print("There is no receiver device attached")
    #         return
    #
    #     while self.receiver.getQueueLength() > 0:
    #         message_received = self.receiver.getData().decode('utf-8')
    #         # print(f'{self.name}: I have received this message {message_received}')
    #         self.receiver.nextPacket()






import abc
from flockai.interfaces.robot import IRobot
import math
from flockai.utils.graphics import Graphics


class IDrone(IRobot, abc.ABC):
    """
    Define all actions needed for flying a drone and declare the get_input() abstract method
    """
    def __init__(self):
        super().__init__()
        self.basic_time_step = int(self.getBasicTimeStep())
        self.name = self.getName()
        self._set_constants()
        self._set_variables()

    def _set_variables(self):
        """
        Set variables needed for controlling the drone on air
        :return:
        """
        self.target_altitude = 1.0      # The target altitude. Can be modified

    def _set_constants(self):
        """
        Set constants needed for controlling the drone on air
        :return: None
        """
        self.K_VERTICAL_THRUST = 68.5   # this thrust lifts the drone
        self.K_VERTICAL_OFFSET = 0.6    # Vertical offset where the robot actually targets to stabilize itself
        self.K_VERTICAL_P = 3.0         # P constant of the vertical PID
        self.K_ROLL_P = 50.0            # P constant of the roll PID
        self.K_PITCH_P = 30.0           # P constant of the pitch PID
        self.K_YAW_P = 20.0             # P constant of the yaw PID

    def actuate(self):
        """
        Fly the drone by stabilizing it based on camera motor sensors and input
        :return: None
        """
        roll, pitch, yaw = self.imu.getRollPitchYaw()
        roll += math.pi / 2

        altitude = self.gps.getValues()[1]
        roll_acceleration, pitch_acceleration, yaw_acceleration = self.gyro.getValues()

        # Set camera motors based on current gyro data
        self.camera_roll_motor.setPosition(-0.115 * roll_acceleration)
        self.camera_pitch_motor.setPosition(-0.1 * pitch_acceleration)

        # Get input
        roll_disturbance, pitch_disturbance, yaw_disturbance = self.get_input()

        roll_input = self.K_ROLL_P * Graphics.clamp(roll, -1.0, 1.0) + roll_acceleration + roll_disturbance
        pitch_input = self.K_PITCH_P * Graphics.clamp(pitch, -1.0, 1.0) - pitch_acceleration + pitch_disturbance
        yaw_input = yaw_disturbance
        clamped_difference_altitude = Graphics.clamp(self.target_altitude - altitude + self.K_VERTICAL_OFFSET, -1.0,
                                                     1.0)
        vertical_input = self.K_VERTICAL_P * pow(clamped_difference_altitude, 3.0)

        # Actuate based on inputs
        front_left_motor_input = self.K_VERTICAL_THRUST + vertical_input - roll_input - pitch_input + yaw_input
        front_right_motor_input = self.K_VERTICAL_THRUST + vertical_input + roll_input - pitch_input - yaw_input
        rear_left_motor_input = self.K_VERTICAL_THRUST + vertical_input - roll_input + pitch_input - yaw_input
        rear_right_motor_input = self.K_VERTICAL_THRUST + vertical_input + roll_input + pitch_input + yaw_input

        # print(front_left_motor_input, front_right_motor_input, rear_left_motor_input, rear_right_motor_input)
        self.front_left_motor.setVelocity(front_left_motor_input)
        self.front_right_motor.setVelocity(-front_right_motor_input)
        self.rear_left_motor.setVelocity(-rear_left_motor_input)
        self.rear_right_motor.setVelocity(rear_right_motor_input)

    @abc.abstractmethod
    def get_input(self) -> tuple:
        """
        Get the input for actuating the drone
        :return: A tuple with pitch, roll, yaw disturbance to actuate the drone
        """
        raise NotImplementedError



